import os, errno
from eunuchs.socketpair import socketpair
from socket import AF_UNIX, SOCK_STREAM, SOL_SOCKET
from sendmsg import recvmsg, SCM_RIGHTS
from struct import unpack

from twisted.internet import abstract, reactor, defer, protocol, threads, utils, fdesc, error
from twisted.python import log, failure

FUSE_MAX_IN = 8192

class ReceiveFuseFD:
    def __init__(self, sockfd, d):
        self.sockfd = sockfd
        self.d = d

    def fileno(self):
        return self.sockfd

    def doRead(self):
        try:
            message, flags, ancillary = recvmsg(self.sockfd)
        except IOError, ioe:
            if ioe.args[0] == errno.EINTR:
                return
            else:
                self.d.errback(ioe)
        else:
            try:
                level, type, fd = ancillary[0]
                assert(level == SOL_SOCKET)
                assert(type == SCM_RIGHTS)
                fd, = unpack('=i', fd)
            except:
                self.d.errback()
            else:
                self.d.callback(fd)
            reactor.removeReader(self)

    def connectionLost(self, reason):
        pass

    def logPrefix(self):
        return ''

class FuseDevice(abstract.FileDescriptor):
    def __init__(self, protocol):
        abstract.FileDescriptor.__init__(self)
        self.protocol = protocol
        self.mounted = False
        self.state = 'disconnected'

    def connect(self, mnt):
        if self.state != 'disconnected':
            raise ValueError, 'not in disconnected state'
        self.state = 'connecting'
        self.mount_point = mnt
        rfd, sfd = self.fusermount_fds = socketpair(AF_UNIX, SOCK_STREAM)
        newenv = os.environ.copy()
        newenv['_FUSE_COMMFD'] = str(sfd)
        reactor.spawnProcess(protocol.ProcessProtocol(), '/usr/bin/fusermount', args = ('fusermount', mnt,), env = newenv,
                childFDs = {0: 'w', 1: 'r', 2: 'r', sfd: sfd})
        d = defer.Deferred()
        rffd = ReceiveFuseFD(rfd, d)
        reactor.addReader(rffd)
        d.addBoth(self._cbCleanup)
        d.addCallbacks(self._cbFD, self._ebFD)
        return d

    def _cbCleanup(self, res):
        for fd in self.fusermount_fds:
            os.close(fd)
        del self.fusermount_fds
        return res
    
    def _cbFD(self, fd):
        self.fd = fd
        fdesc.setNonBlocking(fd)
        reactor.addReader(self)
        self.mounted = True
        self.disco_trigger = reactor.addSystemEventTrigger('before', 'shutdown', self.loseConnection)
        self.state = 'connected'
        self.protocol.makeConnection(self)

    def _ebFD(self, err):
        self.state = 'disconnected'
        log.msg('fuse: connection failed', err)

    def loseConnection(self):
        if self.state != 'connected':
            raise ValueError, 'not in connected state'
        self.state = 'disconnecting'
        reactor.removeReader(self)
        reactor.removeSystemEventTrigger(self.disco_trigger)
        d = utils.getProcessValue('/usr/bin/fusermount', ('-u', '-q', '-z', self.mount_point))
        d.addBoth(self._cbDisconnect)
        return d

    def _cbDisconnect(self, val):
        self.state = 'disconnected'
        self.protocol.connectionLost(error.ConnectionLost())
        if isinstance(val, failure.Failure):
            log.msg('fuse: failed to disconnect', val)
            return -1
        return val

    def writeSequence(self, iovec):
        # TODO: /dev/fuse supports writev
        return self.write(''.join(iovec))

    def write(self, data):
        try:
            log.msg('fuse: outgoing packet', repr(data))
            os.write(self.fileno(), data)
        except (OSError, IOError), ioe:
            if ioe.args[0] != errno.ENOENT:
                log.msg('fuse: error %s writing to device' % (ioe.args[0],))
            return -ioe.args[0]
        else:
            return 0

    def doRead(self):
        # TODO: /dev/fuse supports readv (split into in_header and arg?)
        try:
            output = os.read(self.fileno(), FUSE_MAX_IN)
        except (OSError, IOError), ioe:
            if ioe.args[0] in [errno.EINTR, errno.ENOENT]:
                return
            if ioe.args[0] != errno.ENODEV:
                log.msg('fuse: error %s reading device' % (ioe.args[0],))
            self.loseConnection()
            return
        # no EOF on this
        self.protocol.dataReceived(output)

    def fileno(self):
        return self.fd

