from struct import pack, unpack
from stat import *
from weakref import WeakValueDictionary
from cStringIO import StringIO
import errno

from twisted.python import log

FUSE_ROOT_ID = 1
FUSE_NAME_MAX = 1024
FUSE_MAX_PATH = 4096

ENTRY_REVALIDATE_TIME = 1
ATTR_REVALIDATE_TIME = 1

fuse_entry_out_format = '=QQQQII'
fuse_attr_format = '=QQQQQQIIIIIIII'
stat_fields = ['st_ino', 'st_size', 'st_blocks', 'st_atime', 'st_mtime', 'st_ctime',
               'st_atimensec', 'st_mtimensec', 'st_ctimensec', 'st_mode', 'st_nlink', 'st_uid', 'st_gid', 'st_rdev']

def copy_and_fill_stat(stat):
    return dict([(k, long(getattr(stat, k, 0))) for k in stat_fields])

def pack_stat(stat):
    return pack(fuse_attr_format, *[stat[k] for k in stat_fields])

class FileSystem:
    def __init__(self, proto):
        root = Node(FUSE_ROOT_ID, 0, 0, '/')
        self.root = root
#        self.id_table = {FUSE_ROOT_ID: root}
        self.id_table = WeakValueDictionary()
        self.id_table[FUSE_ROOT_ID] = root
        self.name_table = {}
        self.proto = proto
        self.ctr = 0
        self.generation = 0
        self.open_dirs = {}

    def next_id(self):
        while True:
            self.ctr += 1
            if self.ctr:
                self.generation += 1
            if self.ctr != 0 and self.ctr not in self.id_table:
                break
        return self.ctr

    def lookup_path(self, nodeid, version, name, path):
        res, stat = self.proto.getattr(path)
        if not res:
            stat = copy_and_fill_stat(stat)
            node = self.find_node(nodeid, name, stat, version)
            arg = pack(fuse_entry_out_format, node.nodeid, node.gen, ENTRY_REVALIDATE_TIME, 0,
                       ATTR_REVALIDATE_TIME, 0) + pack_stat(stat)
            return 0, arg, node.nodeid
        else:
            return res, '', None

    def find_node(self, parent, name, stat, version):
        node = self.name_table.get((parent, name))
        if node is None:
            node = Node(self.next_id(), self.generation, parent, name)
            self.id_table[node.nodeid] = node
            self.name_table[(parent, name)] = node
        # TODO: don't do this if flag FUSE_USE_INO is set
        stat['st_ino'] = node.nodeid
        node.version = version
        return node

    def get_path(self, nodeid, name = ''):
        accum = []
        if name:
            accum.append(name)
        print self.id_table.keys()
        cur = self.id_table.get(nodeid)
        if cur is None:
            log.msg('fuse: failed to look up node %d' % (nodeid,))
            return None
        while cur.nodeid != FUSE_ROOT_ID:
            accum.append(cur.name)
            cur = self.id_table.get(cur.parent)
            if cur is None:
                log.msg('fuse: failed to look up node %d' % (nodeid,))
                return None
        accum.reverse()
        ret = '/' + '/'.join(accum)
        if len(ret) >= FUSE_MAX_PATH:
            log.msg('fuse: path too long: %s' % (ret,))
            return None
        return ret

    def forget_node(self, nodeid, version):
        node = self.id_table[nodeid]
        if node.version == version and nodeid != FUSE_ROOT_ID:
            node.version = 0
            del self.name_table[(node.parent, node.name)]

    def opendir(self):
        od = OpenedDir()
        self.open_dirs[id(od)] = od
        return id(od)

    def fill_dir(self, handle, name, type, ino = 0):
        od = self.open_dirs[handle]
        name = name[:FUSE_NAME_MAX]
        # TODO: use ino instead of -1 if flag FUSE_USE_INO is set
        data = pack('=QII', 0xFFFFFFFFL, len(name), type) + name
        od.io.write(data)
        # padding (gagaga retarded)
        od.io.write('\x00' * (8 - (len(data) % 8)))

    def readdir(self, nodeid, handle, offset, size):
        od = self.open_dirs[handle]
        if not od.filled:
            res = -errno.ENOENT
            path = self.get_path(nodeid)
            if path is not None:
                res = -errno.ENOSYS
                if hasattr(self.proto, 'getdir'):
                    res = self.proto.getdir(path, handle, self.fill_dir)
            if res:
                return res, ''
            od.filled = True
            log.msg('fuse: readdir dump', repr(od.io.getvalue()))
        od.io.seek(offset)
        ret = od.io.read(size)
        return 0, ret

    def releasedir(self, handle, _):
        del self.open_dirs[handle]

    def release(self, nodeid, handle, flags):
        node = self.id_table[nodeid]
        assert(node.open_count > 0)
        node.open_count -= 1
        unlink_hidden = node.is_hidden and not node.open_count
        path = self.get_path(nodeid)
        if hasattr(self.proto, 'release'):
            self.proto.release(path or '-', handle, flags)
        if unlink_hidden and path:
            self.proto.unlink(path)
    
    def read(self, nodeid, handle, offset, size):
        data = ''
        res = -errno.ENOENT
        path = self.get_path(nodeid)
        if path is not None:
            res = -errno.ENOSYS
            if hasattr(self.proto, 'read'):
                res, data = self.proto.read(path, size, offset, handle)
        return res, data

class OpenedDir:
    def __init__(self):
        self.io = StringIO()
        self.filled = False

class Node:
    def __init__(self, nodeid, gen, parent, name):
        self.nodeid = nodeid
        self.gen = gen
        self.parent = parent
        self.name = name
        self.open_count = 0
        self.is_hidden = 0

