from struct import pack, unpack, calcsize
import errno
from cStringIO import StringIO

from twisted.internet import reactor, protocol
from twisted.python import log

import filesys

FUSE_KERNEL_VERSION = 5
FUSE_KERNEL_MINOR_VERSION = 1

FUSE_LOOKUP = 1
FUSE_FORGET = 2
FUSE_GETATTR = 3
FUSE_SETATTR = 4
FUSE_READLINK = 5
FUSE_SYMLINK = 6
FUSE_MKNOD = 8
FUSE_MKDIR = 9
FUSE_UNLINK = 10
FUSE_RMDIR = 11
FUSE_RENAME = 12
FUSE_LINK = 13
FUSE_OPEN = 14
FUSE_READ = 15
FUSE_WRITE = 16
FUSE_STATFS = 17
FUSE_RELEASE = 18
FUSE_FSYNC = 20
FUSE_SETXATTR = 21
FUSE_GETXATTR = 22
FUSE_LISTXATTR = 23
FUSE_REMOVEXATTR = 24
FUSE_FLUSH = 25
FUSE_INIT = 26
FUSE_OPENDIR = 27
FUSE_READDIR = 28
FUSE_RELEASEDIR = 29

dispatch = {FUSE_LOOKUP: 'lookup',
            FUSE_FORGET: 'forget',
            FUSE_GETATTR: 'getattr',
            FUSE_OPEN: 'open',
            FUSE_READ: 'read',
            FUSE_RELEASE: 'release',
            FUSE_INIT: 'init',
            FUSE_OPENDIR: 'opendir',
            FUSE_READDIR: 'readdir',
            FUSE_RELEASEDIR: 'releasedir',
           }

dispatch_reverse = {}
for k,v in dispatch.items():
    dispatch_reverse[v] = k

in_format = '=IIQQIII'
in_format_len = calcsize(in_format)

LEN, OPCODE, UNIQUE, NODEID, UID, GID, PID = range(7)

out_format = '=IiQ'
out_format_len = calcsize(out_format)

class FuseProtocol(protocol.Protocol):
    def __init__(self):
        self.got_init = False
        self.fs = filesys.FileSystem(self)

    def connectionMade(self):
        pass

    def dataReceived(self, data):
        #len, opcode, unique, nodeid, uid, gid, pid
        decode = unpack(in_format, data[:in_format_len])
        method = getattr(self, 'handle_%s' % (dispatch.get(decode[OPCODE], 'default'),), self.handle_default)
        log.msg('fuse: incoming packet', dispatch.get(decode[OPCODE], 'default'), decode, repr(data[in_format_len:]))
        if not self.got_init and decode[OPCODE] != FUSE_INIT:
            self.send_reply(-errno.EPROTO, decode[UNIQUE])
        else:
            method(decode, data[in_format_len:])

    def handle_lookup(self, decode, name):
        name = name.rstrip('\x00') # NULL terminator (why the fuck does FUSE decide to pass this?)
        arg_out = ''
        res = -errno.ENOENT
        path = self.fs.get_path(decode[NODEID], name)
        if path is not None:
            res = -errno.ENOSYS
            if hasattr(self, 'getattr'):
                res, arg_out, nodeid = self.fs.lookup_path(decode[NODEID], decode[UNIQUE], name, path)
        res2 = self.send_reply(res, decode[UNIQUE], arg_out)
        if res == 0 and res2 == -errno.ENOENT:
            self.fs.forget_node(nodeid, decode[UNIQUE])

    def handle_forget(self, decode, arg):
        version, = unpack('=Q', arg)
        self.fs.forget_node(decode[NODEID], version)

    def handle_getattr(self, decode, _):
        res = -errno.ENOENT
        path = self.fs.get_path(decode[NODEID])
        arg_out = ''
        if path is not None:
            res = -errno.ENOSYS
            if hasattr(self, 'getattr'):
                res, stat = self.getattr(path)
        if not res:
            stat = filesys.copy_and_fill_stat(stat)
            # TODO: don't do this if flag FUSE_USE_INO is set
            stat['st_ino'] = decode[NODEID]
            arg_out = pack('=QII', filesys.ATTR_REVALIDATE_TIME, 0, 0) + filesys.pack_stat(stat)
        self.send_reply(res, decode[UNIQUE], arg_out)

    def handle_open(self, decode, arg):
        # XXX: guh how do I abstract this to filesys if it calls send_reply in the middle
        flags, = unpack('=I', arg)
        res = -errno.ENOENT
        path = self.fs.get_path(decode[NODEID])
        if path is not None:
            res = -errno.ENOSYS
            if hasattr(self, 'open'):
                res, handle = self.open(path, flags)
        if not res:
            res2 = self.send_reply(res, decode[UNIQUE], pack('=QI', handle, 0))
            if res2 == -errno.ENOENT:
                if hasattr(self, 'release'):
                    self.release(path, handle, flags)
            else:
                self.fs.id_table[decode[NODEID]].open_count += 1
        else:
            self.send_reply(res, decode[UNIQUE])

    def handle_read(self, decode, arg):
        res, arg_out = self.fs.read(decode[NODEID], *unpack('=QQI', arg))
        self.send_reply(res, decode[UNIQUE], arg_out)

    def handle_release(self, decode, arg):
        self.fs.release(decode[NODEID], *unpack('=QI', arg))
        self.send_reply(0, decode[UNIQUE])

    def handle_init(self, decode, _):
        self.got_init = True
        self.send_reply(0, decode[UNIQUE], pack('=II', FUSE_KERNEL_VERSION, FUSE_KERNEL_MINOR_VERSION))

    def handle_opendir(self, decode, _):
        handle = self.fs.opendir()
        self.send_reply(0, decode[UNIQUE], pack('=qI', handle, 0))

    def handle_readdir(self, decode, arg):
        res, arg_out = self.fs.readdir(decode[NODEID], *unpack('=qQI', arg))
        self.send_reply(res, decode[UNIQUE], arg_out)

    def handle_releasedir(self, decode, arg):
        self.fs.releasedir(*unpack('=qI', arg))
        self.send_reply(0, decode[UNIQUE])

    def handle_default(self, decode, _):
        self.send_reply(-errno.ENOSYS, decode[UNIQUE])

    def send_reply(self, error, unique, arg = ''):
        log.msg('fuse: outgoing arg', unique, error, repr(arg))
        return self.transport.writeSequence([pack('=IiQ', out_format_len+len(arg), error, unique), arg])

    def connectionLost(self, reason):
        pass

